/*******************************************************************************
 * Copyright (c) 2000, 2005 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *******************************************************************************/
module dwt.dnd.ClipboardProxy;

import dwt.dwthelper.utils;

 
import dwt.DWT;
import dwt.internal.Callback;
import dwt.internal.Converter;
import dwt.internal.gtk.GtkSelectionData;
import dwt.internal.gtk.GtkTargetEntry;
import dwt.internal.gtk.OS;
import dwt.widgets.Display;
import dwt.widgets.Event;
import dwt.widgets.Listener;

class ClipboardProxy {
    /* Data is not flushed to the clipboard immediately.
     * This class will remember the data and provide it when requested. 
     */
    Object[] clipboardData;
    Transfer[] clipboardDataTypes;
    Object[] primaryClipboardData;
    Transfer[] primaryClipboardDataTypes;

    Display display;
    Clipboard activeClipboard = null;
    Clipboard activePrimaryClipboard = null;
    Callback getFunc;
    Callback clearFunc;
    
    static String ID = "CLIPBOARD PROXY OBJECT"; //$NON-NLS-1$

static ClipboardProxy _getInstance(final Display display) {
    ClipboardProxy proxy = (ClipboardProxy) display.getData(ID);
    if (proxy !is null) return proxy;
    proxy = new ClipboardProxy(display);
    display.setData(ID, proxy);
    display.addListener(DWT.Dispose, new Listener() {
        public void handleEvent(Event event) {
            ClipboardProxy clipbordProxy = (ClipboardProxy)display.getData(ID);
            if (clipbordProxy is null) return;
            display.setData(ID, null);
            clipbordProxy.dispose();
        }
    });
    return proxy;
}   

ClipboardProxy(Display display) {   
    this.display = display;
    getFunc = new Callback( this, "getFunc", 4); //$NON-NLS-1$
    if (getFunc.getAddress() is 0) DWT.error(DWT.ERROR_NO_MORE_CALLBACKS);
    clearFunc = new Callback( this, "clearFunc", 2); //$NON-NLS-1$
    if (clearFunc.getAddress() is 0) DWT.error(DWT.ERROR_NO_MORE_CALLBACKS);
}

void clear (Clipboard owner, int clipboards) {
    if ((clipboards & DND.CLIPBOARD) !is 0 && activeClipboard is owner) {
        OS.gtk_clipboard_clear(Clipboard.GTKCLIPBOARD);
    }
    if ((clipboards & DND.SELECTION_CLIPBOARD) !is 0 && activePrimaryClipboard is owner) {
        OS.gtk_clipboard_clear(Clipboard.GTKPRIMARYCLIPBOARD);
    }
}

int /*long*/ clearFunc(int /*long*/ clipboard,int /*long*/ user_data_or_owner){
    if (clipboard is Clipboard.GTKCLIPBOARD) {
        activeClipboard = null;
        clipboardData = null;
        clipboardDataTypes = null;
    }
    if (clipboard is Clipboard.GTKPRIMARYCLIPBOARD) {
        activePrimaryClipboard = null;
        primaryClipboardData = null;
        primaryClipboardDataTypes = null;
    }
    return 1;
}

void dispose () {
    if (display is null) return;
    if (activeClipboard !is null) OS.gtk_clipboard_clear(Clipboard.GTKCLIPBOARD);
    if (activePrimaryClipboard !is null) OS.gtk_clipboard_clear(Clipboard.GTKPRIMARYCLIPBOARD);
    display = null;
    if (getFunc !is null ) getFunc.dispose();
    getFunc = null;
    if (clearFunc !is null) clearFunc.dispose();
    clearFunc = null;
    clipboardData = null;
    clipboardDataTypes = null;
    primaryClipboardData = null;
    primaryClipboardDataTypes = null;
}

/**
 * This function provides the data to the clipboard on request.
 * When this clipboard is disposed, the data will no longer be available.
 */
int /*long*/ getFunc(int /*long*/ clipboard, int /*long*/ selection_data, int /*long*/ info, int /*long*/ user_data_or_owner){
    if (selection_data is 0) return 0;
    GtkSelectionData selectionData = new GtkSelectionData();
    OS.memmove(selectionData, selection_data, GtkSelectionData.sizeof);
    TransferData tdata = new TransferData();
    tdata.type = selectionData.target;
    Transfer[] types = (clipboard is Clipboard.GTKCLIPBOARD) ? clipboardDataTypes : primaryClipboardDataTypes;
    int index = -1;
    for (int i = 0; i < types.length; i++) {
        if (types[i].isSupportedType(tdata)) {
            index = i;
            break;
        }
    }
    if (index is -1) return 0;
    Object[] data = (clipboard is Clipboard.GTKCLIPBOARD) ? clipboardData : primaryClipboardData;
    types[index].javaToNative(data[index], tdata);
    if (tdata.format < 8 || tdata.format % 8 !is 0) {
        return 0;
    }
    OS.gtk_selection_data_set(selection_data, tdata.type, tdata.format, tdata.pValue, tdata.length);    
    OS.g_free(tdata.pValue);
    return 1;
}

bool setData(Clipboard owner, Object[] data, Transfer[] dataTypes, int clipboards) { 
    GtkTargetEntry[] entries = new  GtkTargetEntry [0];
    int /*long*/ pTargetsList = 0;
    try {
        for (int i = 0; i < dataTypes.length; i++) {
            Transfer transfer = dataTypes[i];
            int[] typeIds = transfer.getTypeIds();
            String[] typeNames = transfer.getTypeNames();
            for (int j = 0; j < typeIds.length; j++) {
                GtkTargetEntry  entry = new GtkTargetEntry();                       
                entry.info = typeIds[j];
                byte[] buffer = Converter.wcsToMbcs(null, typeNames[j], true);
                int /*long*/ pName = OS.g_malloc(buffer.length);
                OS.memmove(pName, buffer, buffer.length);
                entry.target = pName;
                GtkTargetEntry[] tmp = new GtkTargetEntry [entries.length + 1];
                System.arraycopy(entries, 0, tmp, 0, entries.length);
                tmp[entries.length] = entry;
                entries = tmp;              
            }   
        }
        
        pTargetsList = OS.g_malloc(GtkTargetEntry.sizeof * entries.length);
        int offset = 0;
        for (int i = 0; i < entries.length; i++) {
            OS.memmove(pTargetsList + offset, entries[i], GtkTargetEntry.sizeof);
            offset += GtkTargetEntry.sizeof;
        }
        if ((clipboards & DND.CLIPBOARD) !is 0) {
            if (activeClipboard !is null) OS.gtk_clipboard_clear(Clipboard.GTKCLIPBOARD);
            clipboardData = data;
            clipboardDataTypes = dataTypes;
            int /*long*/ getFuncProc = getFunc.getAddress();
            int /*long*/ clearFuncProc = clearFunc.getAddress();
            if (!OS.gtk_clipboard_set_with_data(Clipboard.GTKCLIPBOARD, pTargetsList, entries.length, getFuncProc, clearFuncProc, 0)) {
                return false;
            }
            activeClipboard = owner;
        }
        if ((clipboards & DND.SELECTION_CLIPBOARD) !is 0) {
            if (activePrimaryClipboard !is null) OS.gtk_clipboard_clear(Clipboard.GTKPRIMARYCLIPBOARD);
            primaryClipboardData = data;
            primaryClipboardDataTypes = dataTypes;
            int /*long*/ getFuncProc = getFunc.getAddress();
            int /*long*/ clearFuncProc = clearFunc.getAddress();
            if (!OS.gtk_clipboard_set_with_data(Clipboard.GTKPRIMARYCLIPBOARD, pTargetsList, entries.length, getFuncProc, clearFuncProc, 0)) {
                return false;
            }
            activePrimaryClipboard = owner;
        }
        return true;
    } finally {
        for (int i = 0; i < entries.length; i++) {
            GtkTargetEntry entry = entries[i];
            if( entry.target !is 0) OS.g_free(entry.target);
        }
        if (pTargetsList !is 0) OS.g_free(pTargetsList);
    }
}
}
